import logging
import multiprocessing
import threading
import time

import workflows
import workflows.frontend.utilization
import workflows.services
import workflows.transport
import workflows.util
from workflows.services.common_service import CommonService

basestring = (str, bytes)


class Frontend:
    """The frontend class encapsulates the actual service. It controls the
    service process and keeps the connection to the transport layer. It
    can process control messages directly, or pass messages on to the
    service.
    """

    def __init__(
        self,
        transport=None,
        service=None,
        transport_command_channel=None,
        restart_service=False,
        verbose_service=False,
        environment=None,
    ):
        """Create a frontend instance. Connect to the transport layer, start any
        requested service, begin broadcasting status information and listen
        for control commands.
        :param restart_service:
            If the service process dies unexpectedly the frontend should start
            a new instance.
        :param service:
            A class or name of the class to be instantiated in a subprocess as
            service.
        :param transport:
            Either the name of a transport class, a transport class, or a
            transport class object.
        :param transport_command_channel:
            An optional channel of a transport subscription to be listened to for
            commands.
        :param verbose_service:
            If set, run services with increased logging level (DEBUG).
        :param environment:
            An optional dictionary that is passed to started services.
        """
        self.__lock = threading.RLock()
        self.__hostid = workflows.util.generate_unique_host_id()
        self._service = None  # pointer to the service instance
        self._service_class_name = None
        self._service_factory = None  # pointer to the service class
        self._service_name = None
        self._service_starttime = None
        self._service_rapidstarts = None
        self._pipe_commands = None  # frontend -> service
        self._pipe_service = None  # frontend <- service
        self._service_status = CommonService.SERVICE_STATUS_NONE
        self._service_status_announced = CommonService.SERVICE_STATUS_NONE

        self.restart_service = restart_service
        self.shutdown = False

        # Status broadcast related variables
        self._status_interval = 6
        self._status_last_broadcast = 0
        self._status_idle_since = None
        self._utilization = workflows.frontend.utilization.UtilizationStatistics(
            summation_period=self._status_interval
        )

        # Set up logging
        self._verbose_service = verbose_service

        class LogAdapter:
            """A helper class that acts like a dictionary, but actually reads its
            values from the get_status() function."""

            status_fn = self.get_status
            status = status_fn()

            def __iter__(self):
                """Update cached status values, renaming the keys for logging.
                Return a dictionary key iterator."""
                self.status = {"workflows_" + k: v for k, v in self.status_fn().items()}
                return self.status.__iter__()

            def __getitem__(self, key):
                """Return a value from the status dictionary."""
                return self.status.__getitem__(key)

        self.log = logging.LoggerAdapter(
            logging.getLogger("workflows.frontend"), LogAdapter()
        )
        self.log.warn = self.log.warning  # Add support for deprecated .warn

        # Connect to the network transport layer
        if transport is None or isinstance(transport, basestring):
            self._transport_factory = workflows.transport.lookup(transport)
        else:
            self._transport_factory = transport
        assert hasattr(
            self._transport_factory, "__call__"
        ), "No valid transport factory given"
        self._transport = self._transport_factory()
        assert self._transport.connect(), "Could not connect to transport layer"

        if transport_command_channel:
            self._transport.subscribe_broadcast(
                transport_command_channel, self.process_transport_command
            )
            self.log.debug("Listening for commands on transport layer")

        # Save environment for service starts
        self._service_environment = environment

        # Start initial service if one has been requested
        self._service_factory = service
        if service is not None:
            self.update_status(CommonService.SERVICE_STATUS_NEW)
            self.switch_service()
        else:
            self.update_status()

    def update_status(self, status_code=None):
        """Update the service status kept inside the frontend (_service_status).
        The status is broadcast over the network immediately. If the status
        changes to IDLE then this message is delayed. The IDLE status is only
        broadcast if it is held for over 0.5 seconds.
        When the status does not change it is still broadcast every
        _status_interval seconds.
        :param status_code: Either an integer describing the service status
                            (see workflows.services.common_service), or None
                            if the status is unchanged.
        """
        if status_code is not None:
            self._service_status = status_code
            self._utilization.update_status(status_code)

        # Check whether IDLE status should be delayed
        if self._service_status == CommonService.SERVICE_STATUS_IDLE:
            if self._status_idle_since is None:
                self._status_idle_since = time.time()
                return
            elif self._status_idle_since + 0.5 > time.time():
                return
        else:
            self._status_idle_since = None

        new_status = self._service_status != self._service_status_announced
        if (
            (
                new_status
                or self._status_last_broadcast + self._status_interval <= time.time()
            )
            and self._transport
            and self._transport.is_connected()
        ):
            self._service_status_announced = self._service_status
            self._transport.broadcast_status(self.get_status())
            self._status_last_broadcast = time.time()

    def run(self):
        """The main loop of the frontend. Here incoming messages from the service
        are processed and forwarded to the corresponding callback methods."""
        self.log.debug("Entered main loop")
        while not self.shutdown:
            # If no service is running slow down the main loop
            if not self._pipe_service:
                time.sleep(0.3)
            self.update_status()
            # While a service is running, check for incoming messages from that service
            if self._pipe_service and self._pipe_service.poll(1):
                try:
                    message = self._pipe_service.recv()
                    if isinstance(message, dict) and "band" in message:
                        # only dictionaries with 'band' entry are valid messages
                        try:
                            handler = getattr(self, "parse_band_" + message["band"])
                        except AttributeError:
                            handler = None
                            self.log.warning("Unknown band %s", str(message["band"]))
                        if handler:
                            #              try:
                            handler(message)
                    #              except Exception:
                    #                print('Uh oh. What to do.')
                    else:
                        self.log.warning("Invalid message received %s", str(message))
                except EOFError:
                    # Service has gone away
                    error_message = False
                    if self._service_status == CommonService.SERVICE_STATUS_END:
                        self.log.info("Service terminated")
                    elif self._service_status == CommonService.SERVICE_STATUS_ERROR:
                        error_message = "Service terminated with error code"
                    elif self._service_status in (
                        CommonService.SERVICE_STATUS_NONE,
                        CommonService.SERVICE_STATUS_NEW,
                        CommonService.SERVICE_STATUS_STARTING,
                    ):
                        error_message = (
                            "Service may have died unexpectedly in "
                            + "initialization (last known status: %s)"
                            % CommonService.human_readable_state.get(
                                self._service_status, self._service_status
                            )
                        )
                    else:
                        error_message = (
                            "Service may have died unexpectedly"
                            " (last known status: %s)"
                            % CommonService.human_readable_state.get(
                                self._service_status, self._service_status
                            )
                        )
                    if error_message:
                        self.log.error(error_message)
                    self._terminate_service()
                    if self.restart_service:
                        self.exponential_backoff()
                    else:
                        self.shutdown = True
                        if error_message:
                            raise workflows.Error(error_message)

            with self.__lock:
                if (
                    self._service is None
                    and self.restart_service
                    and self._service_factory
                ):
                    self.update_status(status_code=CommonService.SERVICE_STATUS_NEW)
                    self.switch_service()

            # Check that the transport is alive
            if not self._transport.is_connected():
                self._terminate_service()
                raise workflows.Error("Lost transport layer connection")
        self.log.debug("Left main loop")
        self.update_status(status_code=CommonService.SERVICE_STATUS_TEARDOWN)
        self._terminate_service()
        self.log.debug("Terminating.")

    def send_command(self, command):
        """Send command to service via the command queue."""
        if self._pipe_commands:
            self._pipe_commands.send(command)
        else:
            if self.shutdown:
                # Stop delivering messages in shutdown.
                self.log.info(
                    "During shutdown no command queue pipe found for command\n%s",
                    str(command),
                )
            else:
                self.log.error(
                    "No command queue pipe found for command\n%s", str(command)
                )

    def process_transport_command(self, header, message):
        """Parse a command coming in through the transport command subscription"""
        if not isinstance(message, dict):
            return

        relevant = False
        if "host" in message:  # Filter by host
            if message["host"] != self.__hostid:
                return
            relevant = True
        if "service" in message:  # Filter by service
            if message["service"] != self._service_class_name:
                return
            relevant = True
        if not relevant:  # Ignore message unless at least one filter matches
            return

        if message.get("command"):
            self.log.info(
                "Received command '%s' via transport layer", message["command"]
            )
            if message["command"] == "shutdown":
                self.shutdown = True
        else:
            self.log.warning("Received invalid transport command message")

    def parse_band_log(self, message):
        """Process incoming logging messages from the service."""
        if "payload" in message and hasattr(message["payload"], "name"):
            record = message["payload"]
            for k in dir(record):
                if k.startswith("workflows_exc_"):
                    setattr(record, k[14:], getattr(record, k))
                    delattr(record, k)
            for k, v in self.get_status().items():
                setattr(record, "workflows_" + k, v)
            logging.getLogger(record.name).handle(record)
        else:
            self.log.warning(
                "Received broken record on log band\n" + "Message: %s\nRecord: %s",
                str(message),
                str(
                    hasattr(message.get("payload"), "__dict__")
                    and message["payload"].__dict__
                ),
            )

    def parse_band_request_termination(self, message):
        """Service declares it should be terminated."""
        self.log.debug("Service requests termination")
        self._terminate_service()
        if not self.restart_service:
            self.shutdown = True

    def parse_band_set_name(self, message):
        """Process incoming message indicating service name change."""
        if message.get("name"):
            self._service_name = message["name"]
        else:
            self.log.warning(
                "Received broken record on set_name band\nMessage: %s", str(message)
            )

    def parse_band_status_update(self, message):
        """Process incoming status updates from the service."""
        self.log.debug("Status update: " + str(message))
        self.update_status(status_code=message["statuscode"])

    def get_host_id(self):
        """Get a cached copy of the host id."""
        return self.__hostid

    def get_status(self):
        """Returns a dictionary containing all relevant status information to be
        broadcast across the network."""
        return {
            "host": self.__hostid,
            "status": self._service_status_announced,
            "statustext": CommonService.human_readable_state.get(
                self._service_status_announced
            ),
            "service": self._service_name,
            "serviceclass": self._service_class_name,
            "utilization": self._utilization.report(),
            "workflows": workflows.version(),
        }

    def exponential_backoff(self):
        """A function that keeps waiting longer and longer the more rapidly it is called.
        It can be used to increasingly slow down service starts when they keep failing."""
        last_service_switch = self._service_starttime
        if not last_service_switch:
            return
        time_since_last_switch = time.time() - last_service_switch

        if not self._service_rapidstarts:
            self._service_rapidstarts = 0
        minimum_wait = 0.1 * (2 ** self._service_rapidstarts)
        minimum_wait = min(5, minimum_wait)

        if time_since_last_switch > 10:
            self._service_rapidstarts = 0
            return
        self._service_rapidstarts += 1
        self.log.debug("Slowing down service starts (%.1f seconds)", minimum_wait)
        time.sleep(minimum_wait)

    def switch_service(self, new_service=None):
        """Start a new service in a subprocess.
        :param new_service: Either a service name or a service class. If not set,
                            start up a new instance of the previous class
        :return: True on success, False on failure.
        """
        if new_service:
            self._service_factory = new_service
        with self.__lock:
            # Terminate existing service if necessary
            if self._service is not None:
                self._terminate_service()

            # Find service class if necessary
            if isinstance(self._service_factory, basestring):
                self._service_factory = workflows.services.lookup(self._service_factory)
            if not self._service_factory:
                return False

            # Set up new service object
            service_instance = self._service_factory(
                environment=self._service_environment
            )

            # Set up pipes and connect service object
            svc_commands, self._pipe_commands = multiprocessing.Pipe(False)
            self._pipe_service, svc_tofrontend = multiprocessing.Pipe(False)
            service_instance.connect(commands=svc_commands, frontend=svc_tofrontend)

            # Set up transport layer for new service
            service_instance.transport = self._transport_factory()

            # Start new service in a separate process
            self._service = multiprocessing.Process(
                target=service_instance.start,
                args=(),
                kwargs={"verbose_log": self._verbose_service},
            )
            self._service_name = service_instance.get_name()
            self._service_class_name = service_instance.__class__.__name__
            self._service.daemon = True
            self._service.name = "workflows-service"
            self._service.start()
            self._service_starttime = time.time()

            # Starting the process copies all file descriptors.
            # At this point (and no sooner!) the passed pipe objects must be closed
            # in this process here.
            svc_commands.close()
            svc_tofrontend.close()
        self.log.info("Started service: %s", self._service_name)
        return True

    def _terminate_service(self):
        """Force termination of running service.
        Disconnect queues, end queue feeder threads.
        Wait for service process to clear, drop all references."""
        with self.__lock:
            if self._service:
                self._service.terminate()
            if self._pipe_commands:
                self._pipe_commands.close()
            if self._pipe_service:
                self._pipe_service.close()
            self._pipe_commands = None
            self._pipe_service = None
            self._service_class_name = None
            self._service_name = None
            if self._service_status != CommonService.SERVICE_STATUS_TEARDOWN:
                self.update_status(status_code=CommonService.SERVICE_STATUS_END)
            if self._service:
                self._service.join()  # must wait for process to be actually destroyed
            self._service = None



from optparse import SUPPRESS_HELP, OptionParser

import workflows
import workflows.frontend
import workflows.services
import workflows.transport


class ServiceStarter:
    """A helper class to start a workflows service from the command line.
    A number of hooks are provided so that this class can be subclassed and
    used in a number of scenarios."""

    @staticmethod
    def on_parser_preparation(parser):
        """Plugin hook to manipulate the OptionParser object before command line
        parsing. If a value is returned here it will replace the OptionParser
        object."""

    @staticmethod
    def on_parsing(options, args):
        """Plugin hook to manipulate the command line parsing results.
        A tuple of values can be returned, which will replace (options, args).
        """

    @staticmethod
    def on_transport_factory_preparation(transport_factory):
        """Plugin hook to intercept/manipulate newly created Transport factories
        before first invocation."""

    @staticmethod
    def on_transport_preparation(transport):
        """Plugin hook to intercept/manipulate newly created Transport objects
        before connecting."""

    @staticmethod
    def before_frontend_construction(kwargs):
        """Plugin hook to manipulate the Frontend object constructor arguments. If
        a value is returned here it will replace the keyword arguments
        dictionary passed to the constructor."""

    @staticmethod
    def on_frontend_preparation(frontend):
        """Plugin hook to manipulate the Frontend object before starting it. If a
        value is returned here it will replace the Frontend object."""

    def run(
        self,
        cmdline_args=None,
        program_name="start_service",
        version=workflows.version(),
        **kwargs
    ):
        """Example command line interface to start services.
        :param cmdline_args: List of command line arguments to pass to parser
        :param program_name: Name of the command line tool to display in help
        :param version: Version number to print when run with '--version'
        """

        # Enumerate all known services
        known_services = workflows.services.get_known_services()

        # Set up parser
        parser = OptionParser(
            usage=program_name + " [options]" if program_name else None, version=version
        )
        parser.add_option("-?", action="help", help=SUPPRESS_HELP)
        parser.add_option(
            "-s",
            "--service",
            dest="service",
            metavar="SVC",
            default=None,
            help="Name of the service to start. Known services: "
            + ", ".join(known_services),
        )
        parser.add_option(
            "-t",
            "--transport",
            dest="transport",
            metavar="TRN",
            default="StompTransport",
            help="Transport mechanism. Known mechanisms: "
            + ", ".join(workflows.transport.get_known_transports())
            + " (default: %default)",
        )
        workflows.transport.add_command_line_options(parser)

        # Call on_parser_preparation hook
        parser = self.on_parser_preparation(parser) or parser

        # Parse command line options
        (options, args) = parser.parse_args(cmdline_args)

        # Call on_parsing hook
        (options, args) = self.on_parsing(options, args) or (options, args)

        # Create Transport factory
        transport_factory = workflows.transport.lookup(options.transport)

        # Call on_transport_factory_preparation hook
        transport_factory = (
            self.on_transport_factory_preparation(transport_factory)
            or transport_factory
        )

        # Set up on_transport_preparation hook to affect newly created transport objects
        true_transport_factory_call = transport_factory.__call__

        def on_transport_preparation_hook():
            transport_object = true_transport_factory_call()
            return self.on_transport_preparation(transport_object) or transport_object

        transport_factory.__call__ = on_transport_preparation_hook

        # When service name is specified, check if service exists or can be derived
        if options.service and options.service not in known_services:
            matching = [s for s in known_services if s.startswith(options.service)]
            if not matching:
                matching = [
                    s
                    for s in known_services
                    if s.lower().startswith(options.service.lower())
                ]
            if matching and len(matching) == 1:
                options.service = matching[0]

        kwargs.update({"service": options.service, "transport": transport_factory})

        # Call before_frontend_construction hook
        kwargs = self.before_frontend_construction(kwargs) or kwargs

        # Create Frontend object
        frontend = workflows.frontend.Frontend(**kwargs)

        # Call on_frontend_preparation hook
        frontend = self.on_frontend_preparation(frontend) or frontend

        # Start Frontend
        try:
            frontend.run()
        except KeyboardInterrupt:
            print("\nShutdown via Ctrl+C")


if __name__ == "__main__":  # pragma: no cover
    ServiceStarter().run()



import pkg_resources


def lookup(service):
    """Find a service class based on a name.
    :param service: Name of the service
    :return: A service class
    """
    return get_known_services().get(service)


def get_known_services():
    """Return a dictionary of all known services.
    :return: A dictionary containing entries { service name : service class }

    Future: This will change to a dictionary containing references to
            factories:               { service name : service class factory }
            A factory is a function that takes no arguments and returns an
            uninstantiated service class. This will avoid importing all
            service classes.
    """
    if not hasattr(get_known_services, "cache"):
        setattr(
            get_known_services,
            "cache",
            {
                e.name: e.load()
                for e in pkg_resources.iter_entry_points("workflows.services")
            },
        )
    register = get_known_services.cache.copy()
    return register



import pytest
import workflows.contrib.start_service
from unittest import mock


def test_get_command_line_help(capsys):
    """Running the start_service script with --help should display command line help and exit."""
    with pytest.raises(SystemExit):
        workflows.contrib.start_service.ServiceStarter().run(
            ["--help"], program_name="sentinelvalue"
        )
    out, err = capsys.readouterr()
    assert "Usage: sentinelvalue" in out


@mock.patch("workflows.contrib.start_service.OptionParser")
@mock.patch("workflows.contrib.start_service.workflows.transport.lookup")
@mock.patch("workflows.contrib.start_service.workflows.frontend")
@mock.patch("workflows.contrib.start_service.workflows.services")
def test_script_initialises_transport_and_starts_frontend(
    mock_services, mock_frontend, mock_tlookup, mock_parser
):
    """Check that the start_service script sets up the transport mechanism and the frontend properly.
    Correct service should be selected and the frontend started."""
    mock_options = mock.Mock()
    mock_options.service = "someservice"
    mock_options.transport = mock.sentinel.transport
    mock_parser.return_value.parse_args.return_value = (mock_options, mock.Mock())
    mock_services.get_known_services.return_value = {"SomeService": None}

    workflows.contrib.start_service.ServiceStarter().run(
        cmdline_args=["-s", "some"], version=mock.sentinel.version
    )

    mock_tlookup.assert_called_once_with(mock.sentinel.transport)
    mock_parser.assert_called_once_with(usage=mock.ANY, version=mock.sentinel.version)
    mock_frontend.Frontend.assert_called_once_with(
        service="SomeService", transport=mock_tlookup.return_value
    )
    mock_frontend.Frontend.return_value.run.assert_called_once_with()



import curses
import threading
import time

import workflows.transport
from workflows.services.common_service import CommonService

basestring = (str, bytes)


class Monitor:  # pragma: no cover
    """A sample implementation of a status monitor showing all running services.
    To use this example class, pass in a transport object and call the run()
    method."""

    shutdown = False
    """Set to true to end the main loop and shut down the service monitor."""

    cards = {}
    """Register card shown for seen services"""

    border_chars = ()
    """Characters used for frame borders."""
    border_chars_text = ("|", "|", "=", "=", "/", "\\", "\\", "/")
    """Example alternative set of frame border characters."""

    def __init__(self, transport=None):
        """Set up monitor and connect to the network transport layer"""
        if transport is None or isinstance(transport, basestring):
            self._transport = workflows.transport.lookup(transport)()
        else:
            self._transport = transport()
        assert self._transport.connect(), "Could not connect to transport layer"
        self._lock = threading.RLock()
        self._node_status = {}
        self.message_box = None
        self._transport.subscribe_broadcast(
            "transient.status", self.update_status, retroactive=True
        )

    def update_status(self, header, message):
        """Process incoming status message. Acquire lock for status dictionary before updating."""
        with self._lock:
            if self.message_box:
                self.message_box.erase()
                self.message_box.move(0, 0)
                for n, field in enumerate(header):
                    if n == 0:
                        self.message_box.addstr(field + ":", curses.color_pair(1))
                    else:
                        self.message_box.addstr(
                            ", " + field + ":", curses.color_pair(1)
                        )
                    self.message_box.addstr(header[field])
                self.message_box.addstr(": ", curses.color_pair(1))
                self.message_box.addstr(
                    str(message), curses.color_pair(2) + curses.A_BOLD
                )
                self.message_box.refresh()
            if (
                message["host"] not in self._node_status
                or int(header["timestamp"])
                >= self._node_status[message["host"]]["last_seen"]
            ):
                self._node_status[message["host"]] = message
                self._node_status[message["host"]]["last_seen"] = int(
                    header["timestamp"]
                )

    def run(self):
        """A wrapper for the real _run() function to cleanly enable/disable the
        curses environment."""
        curses.wrapper(self._run)

    def _boxwin(
        self, height, width, row, column, title=None, title_x=7, color_pair=None
    ):
        with self._lock:
            box = curses.newwin(height, width, row, column)
            box.clear()
            if color_pair:
                box.attron(curses.color_pair(color_pair))
            box.border(*self.border_chars)
            if title:
                box.addstr(0, title_x, " " + title + " ")
            if color_pair:
                box.attroff(curses.color_pair(color_pair))
            box.noutrefresh()
            return curses.newwin(height - 2, width - 2, row + 1, column + 1)

    def _redraw_screen(self, stdscr):
        """Redraw screen. This could be to initialize, or to redraw after resizing."""
        with self._lock:
            stdscr.clear()
            stdscr.addstr(
                0, 0, "workflows service monitor -- quit with Ctrl+C", curses.A_BOLD
            )
            stdscr.refresh()
            self.message_box = self._boxwin(
                5, curses.COLS, 2, 0, title="last seen message", color_pair=1
            )
            self.message_box.scrollok(True)
            self.cards = []

    def _get_card(self, number):
        with self._lock:
            if number < len(self.cards):
                return self.cards[number]
            if number == len(self.cards):
                max_cards_horiz = int(curses.COLS / 35)
                self.cards.append(
                    self._boxwin(
                        6,
                        35,
                        7 + 6 * (number // max_cards_horiz),
                        35 * (number % max_cards_horiz),
                        color_pair=3,
                    )
                )
                return self.cards[number]
            raise RuntimeError("Card number too high")

    def _erase_card(self, number):
        """Destroy cards with this or higher number."""
        with self._lock:
            if number < (len(self.cards) - 1):
                self._erase_card(number + 1)
            if number > (len(self.cards) - 1):
                return
            max_cards_horiz = int(curses.COLS / 35)
            obliterate = curses.newwin(
                6,
                35,
                7 + 6 * (number // max_cards_horiz),
                35 * (number % max_cards_horiz),
            )
            obliterate.erase()
            obliterate.noutrefresh()
            del self.cards[number]

    def _run(self, stdscr):
        """Start the actual service monitor"""
        with self._lock:
            curses.use_default_colors()
            curses.curs_set(False)
            curses.init_pair(1, curses.COLOR_RED, -1)
            curses.init_pair(2, curses.COLOR_BLACK, -1)
            curses.init_pair(3, curses.COLOR_GREEN, -1)
            self._redraw_screen(stdscr)

        try:
            while not self.shutdown:
                now = int(time.time())
                with self._lock:
                    overview = self._node_status.copy()
                cardnumber = 0
                for host, status in overview.items():
                    age = now - int(status["last_seen"] / 1000)
                    with self._lock:
                        if age > 90:
                            del self._node_status[host]
                        else:
                            card = self._get_card(cardnumber)
                            card.erase()
                            card.move(0, 0)
                            card.addstr("Host: ", curses.color_pair(3))
                            card.addstr(host)
                            card.move(1, 0)
                            card.addstr("Service: ", curses.color_pair(3))
                            if "service" in status and status["service"]:
                                card.addstr(status["service"])
                            else:
                                card.addstr("---", curses.color_pair(2))
                            card.move(2, 0)
                            card.addstr("State: ", curses.color_pair(3))
                            if "status" in status:
                                status_code = status["status"]
                                state_string = CommonService.human_readable_state.get(
                                    status_code, str(status_code)
                                )
                                state_color = None
                                if status_code in (
                                    CommonService.SERVICE_STATUS_PROCESSING,
                                    CommonService.SERVICE_STATUS_TIMER,
                                ):
                                    state_color = curses.color_pair(3) + curses.A_BOLD
                                if status_code == CommonService.SERVICE_STATUS_IDLE:
                                    state_color = curses.color_pair(2) + curses.A_BOLD
                                if status_code == CommonService.SERVICE_STATUS_ERROR:
                                    state_color = curses.color_pair(1)
                                if state_color:
                                    card.addstr(state_string, state_color)
                                else:
                                    card.addstr(state_string)
                            card.move(3, 0)
                            if age >= 10:
                                card.addstr(
                                    "last seen %d seconds ago" % age,
                                    curses.color_pair(1)
                                    + (0 if age < 60 else curses.A_BOLD),
                                )
                            card.noutrefresh()
                            cardnumber = cardnumber + 1
                if cardnumber < len(self.cards):
                    with self._lock:
                        self._erase_card(cardnumber)
                with self._lock:
                    curses.doupdate()
                time.sleep(0.2)
        except KeyboardInterrupt:
            pass  # User pressed CTRL+C
        self._transport.disconnect()



import pkg_resources

default_transport = "StompTransport"


def lookup(transport):
    """Get a transport layer class based on its name."""
    return get_known_transports().get(
        transport, get_known_transports()[default_transport]
    )


def add_command_line_options(parser):
    """Add command line options for all available transport layer classes."""
    for transport in get_known_transports().values():
        transport().add_command_line_options(parser)


def get_known_transports():
    """Return a dictionary of all known transport mechanisms."""
    if not hasattr(get_known_transports, "cache"):
        setattr(
            get_known_transports,
            "cache",
            {
                e.name: e.load()
                for e in pkg_resources.iter_entry_points("workflows.transport")
            },
        )
    return get_known_transports.cache.copy()



import copy
import json
import string

import workflows

basestring = (str, bytes)


class Recipe:
    """Object containing a processing recipe that can be passed to services.
    A recipe describes how all involved services are connected together, how
    data should be passed and how errors should be handled."""

    recipe = {}
    """The processing recipe is encoded in this dictionary."""
    # TODO: Describe format

    def __init__(self, recipe=None):
        """Constructor allows passing in a recipe dictionary."""
        if isinstance(recipe, basestring):
            self.recipe = self.deserialize(recipe)
        elif recipe:
            self.recipe = self._sanitize(recipe)

    def deserialize(self, string):
        """Convert a recipe that has been stored as serialized json string to a
        data structure."""
        return self._sanitize(json.loads(string))

    @staticmethod
    def _sanitize(recipe):
        """Clean up a recipe that may have been stored as serialized json string.
        Convert any numerical pointers that are stored as strings to integers."""
        recipe = recipe.copy()
        for k in list(recipe):
            if k not in ("start", "error") and int(k) and k != int(k):
                recipe[int(k)] = recipe[k]
                del recipe[k]
        for k in list(recipe):
            if "output" in recipe[k] and not isinstance(
                recipe[k]["output"], (list, dict)
            ):
                recipe[k]["output"] = [recipe[k]["output"]]
            # dicts should be normalized, too
        if "start" in recipe:
            recipe["start"] = [tuple(x) for x in recipe["start"]]
        return recipe

    def serialize(self):
        """Write out the current recipe as serialized json string."""
        return json.dumps(self.recipe)

    def pretty(self):
        """Write out the current recipe as serialized json string with pretty formatting."""
        return json.dumps(self.recipe, indent=2)

    def __getitem__(self, item):
        """Allow direct dictionary access to recipe elements."""
        return self.recipe.__getitem__(item)

    def __contains__(self, item):
        """Testing for presence of recipe elements."""
        return item in self.recipe

    def __eq__(self, other):
        """Overload equality operator (!=) to allow comparing recipe objects
        with one another and with their string representations."""
        if isinstance(other, Recipe):
            return self.recipe == other.recipe
        if isinstance(other, dict):
            return self.recipe == self._sanitize(other)
        return self.recipe == self.deserialize(other)

    def __ne__(self, other):
        """Overload inequality operator (!=) to allow comparing recipe objects
        with one another and with their string representations."""
        result = self.__eq__(other)
        if result is NotImplemented:
            return result
        return not result

    def __hash__(self):
        """Recipe objects are mutable and therefore should not be hashable."""
        return None

    def validate(self):
        """Check whether the encoded recipe is valid. It must describe a directed
        acyclical graph, all connections must be defined, etc."""
        if not self.recipe:
            raise workflows.Error("Invalid recipe: No recipe defined")

        # Without a 'start' node nothing would happen
        if "start" not in self.recipe:
            raise workflows.Error('Invalid recipe: "start" node missing')
        if not self.recipe["start"]:
            raise workflows.Error('Invalid recipe: "start" node empty')
        if not all(
            isinstance(x, (list, tuple)) and len(x) == 2 for x in self.recipe["start"]
        ):
            raise workflows.Error('Invalid recipe: "start" node invalid')
        if any(x[0] == "start" for x in self.recipe["start"]):
            raise workflows.Error('Invalid recipe: "start" node points to itself')

        # Check that 'error' node points to regular nodes only
        if "error" in self.recipe and isinstance(
            self.recipe["error"], (list, tuple, basestring)
        ):
            if "start" in self.recipe["error"]:
                raise workflows.Error(
                    'Invalid recipe: "error" node points to "start" node'
                )
            if "error" in self.recipe["error"]:
                raise workflows.Error('Invalid recipe: "error" node points to itself')

        # All other nodes must be numeric
        nodes = list(
            filter(
                lambda x: not isinstance(x, int) and x not in ("start", "error"),
                self.recipe,
            )
        )
        if nodes:
            raise workflows.Error('Invalid recipe: Node "%s" is not numeric' % nodes[0])

        # Detect cycles
        touched_nodes = {"start", "error"}

        def flatten_links(struct):
            """Take an output/error link object, list or dictionary and return flat list of linked nodes."""
            if struct is None:
                return []
            if isinstance(struct, int):
                return [struct]
            if isinstance(struct, list):
                if not all(isinstance(x, int) for x in struct):
                    raise workflows.Error(
                        "Invalid recipe: Invalid link in recipe (%s)" % str(struct)
                    )
                return struct
            if isinstance(struct, dict):
                joined_list = []
                for sub_list in struct.values():
                    joined_list += flatten_links(sub_list)
                return joined_list
            raise workflows.Error(
                "Invalid recipe: Invalid link in recipe (%s)" % str(struct)
            )

        def find_cycles(path):
            """Depth-First-Search helper function to identify cycles."""
            if path[-1] not in self.recipe:
                raise workflows.Error(
                    'Invalid recipe: Node "%s" is referenced via "%s" but missing'
                    % (str(path[-1]), str(path[:-1]))
                )
            touched_nodes.add(path[-1])
            node = self.recipe[path[-1]]
            for outgoing in ("output", "error"):
                if outgoing in node:
                    references = flatten_links(node[outgoing])
                    for n in references:
                        if n in path:
                            raise workflows.Error(
                                "Invalid recipe: Recipe contains cycle (%s -> %s)"
                                % (str(path), str(n))
                            )
                        find_cycles(path + [n])

        for link in self.recipe["start"]:
            find_cycles(["start", link[0]])
        if "error" in self.recipe:
            if isinstance(self.recipe["error"], (list, tuple)):
                for link in self.recipe["error"]:
                    find_cycles(["error", link])
            else:
                find_cycles(["error", self.recipe["error"]])

        # Test recipe for unreferenced nodes
        for node in self.recipe:
            if node not in touched_nodes:
                raise workflows.Error(
                    'Invalid recipe: Recipe contains unreferenced node "%s"' % str(node)
                )

    def apply_parameters(self, parameters):
        """Recursively apply dictionary entries in 'parameters' to {item}s in recipe
        structure, leaving undefined {item}s as they are. A special case is a
        {$REPLACE:item}, which replaces the string with a copy of the referenced
        parameter item.

        Examples:

        parameters = { 'x':'5' }
        apply_parameters( { '{x}': '{y}' }, parameters )
           => { '5': '{y}' }

        parameters = { 'y':'5' }
        apply_parameters( { '{x}': '{y}' }, parameters )
           => { '{x}': '5' }

        parameters = { 'x':'3', 'y':'5' }
        apply_parameters( { '{x}': '{y}' }, parameters )
           => { '3': '5' }

        parameters = { 'l': [ 1, 2 ] }
        apply_parameters( { 'x': '{$REPLACE:l}' }, parameters )
           => { 'x': [ 1, 2 ] }
        """

        class SafeString:
            def __init__(self, s):
                self.string = s

            def __repr__(self):
                return "{" + self.string + "}"

            def __str__(self):
                return "{" + self.string + "}"

            def __getitem__(self, item):
                return SafeString(self.string + "[" + item + "]")

        class SafeDict(dict):
            """A dictionary that returns undefined keys as {keyname}.
            This can be used to selectively replace variables in datastructures."""

            def __missing__(self, key):
                return SafeString(key)

        # By default the python formatter class is used to resolve {item} references
        formatter = string.Formatter()

        # Special format strings "{$REPLACE:(...)}" use this data structure
        # formatter to return the referenced data structure rather than a formatted
        # string.
        ds_formatter = string.Formatter()

        def ds_format_field(value, spec):
            ds_format_field.last = value
            return ""

        ds_formatter.format_field = ds_format_field

        params = SafeDict(parameters)

        def _recursive_apply(item):
            """Helper function to recursively apply replacements."""
            if isinstance(item, basestring):
                if item.startswith("{$REPLACE") and item.endswith("}"):
                    try:
                        ds_formatter.vformat("{" + item[10:-1] + "}", (), parameters)
                    except KeyError:
                        return None
                    return copy.deepcopy(ds_formatter.format_field.last)
                else:
                    return formatter.vformat(item, (), params)
            if isinstance(item, dict):
                return {
                    _recursive_apply(key): _recursive_apply(value)
                    for key, value in item.items()
                }
            if isinstance(item, tuple):
                return tuple(_recursive_apply(list(item)))
            if isinstance(item, list):
                return [_recursive_apply(x) for x in item]
            return item

        self.recipe = _recursive_apply(self.recipe)

    def merge(self, other):
        """Merge two recipes together, returning a single recipe containing all
        nodes.
        Note: This does NOT yet return a minimal recipe.
        :param other: A Recipe object that should be merged with the current
                      Recipe object.
        :return: A new Recipe object containing information from both recipes.
        """

        # Merging empty values returns a copy of the original
        if not other:
            return Recipe(self.recipe)

        # When a string is passed, merge with a constructed recipe object
        if isinstance(other, basestring):
            return self.merge(Recipe(other))

        # Merging empty recipes returns a copy of the original
        if not other.recipe:
            return Recipe(self.recipe)

        # If own recipe empty, use other recipe
        if not self.recipe:
            return Recipe(other.recipe)

        # Assuming both recipes are valid
        self.validate()
        other.validate()

        # Start from current recipe
        new_recipe = self.recipe

        # Find the maximum index of the current recipe
        max_index = max(1, *filter(lambda x: isinstance(x, int), self.recipe.keys()))
        next_index = max_index + 1

        # Set up a translation table for indices and copy all entries
        translation = {}
        for key, value in other.recipe.items():
            if isinstance(key, int):
                if key not in translation:
                    translation[key] = next_index
                    next_index = next_index + 1
                new_recipe[translation[key]] = value

        # Rewrite all copied entries to point to new keys
        def translate(x):
            if isinstance(x, list):
                return list(map(translate, x))
            elif isinstance(x, tuple):
                return tuple(map(translate, x))
            elif isinstance(x, dict):
                return {k: translate(v) for k, v in x.items()}
            else:
                return translation[x]

        for idx in translation.values():
            if "output" in new_recipe[idx]:
                new_recipe[idx]["output"] = translate(new_recipe[idx]["output"])
            if "error" in new_recipe[idx]:
                new_recipe[idx]["error"] = translate(new_recipe[idx]["error"])

        # Join 'start' nodes
        for (idx, param) in other.recipe["start"]:
            new_recipe["start"].append((translate(idx), param))

        # Join 'error' nodes
        if "error" in other.recipe:
            if "error" not in new_recipe:
                new_recipe["error"] = translate(other.recipe["error"])
            else:
                if isinstance(new_recipe["error"], (list, tuple)):
                    new_recipe["error"] = list(new_recipe["error"])
                else:
                    new_recipe["error"] = list([new_recipe["error"]])
                if isinstance(other.recipe["error"], (list, tuple)):
                    new_recipe["error"].extend(translate(other.recipe["error"]))
                else:
                    new_recipe["error"].append(translate(other.recipe["error"]))

        #   # Minimize DAG
        #   queuehash, topichash = {}, {}
        #   for k, v in new_recipe.items():
        #     if isinstance(v, dict):
        #       if 'queue' in v:
        #         queuehash[v['queue']] = queuehash.get(v['queue'], [])
        #         queuehash[v['queue']].append(k)
        #       if 'topic' in v:
        #         topichash[v['topic']] = topichash.get(v['topic'], [])
        #         topichash[v['topic']].append(k)
        #
        #   print queuehash
        #   print topichash

        return Recipe(new_recipe)



from unittest import mock

import workflows.services
import workflows.services.sample_transaction


def test_services_can_be_looked_up():
    """Attempt to look up the services by their names"""
    service_class = workflows.services.lookup("SampleTxn")
    assert service_class == workflows.services.sample_transaction.SampleTxn
    service_class = workflows.services.lookup("SampleTxnProducer")
    assert service_class == workflows.services.sample_transaction.SampleTxnProducer


def test_txnproducer_registers_idle_timer():
    """Check that the TXN producer registers an idle event handler."""
    p = workflows.services.sample_transaction.SampleTxnProducer()
    mock_idlereg = mock.Mock()
    setattr(p, "_register_idle", mock_idlereg)

    p.initializing()

    mock_idlereg.assert_called_once_with(mock.ANY, p.create_message)


def test_txnproducer_produces_messages():
    """Check that the TXN producer produces messages in the idle event handler."""
    p = workflows.services.sample_transaction.SampleTxnProducer()
    mock_transport = mock.Mock()
    setattr(p, "_transport", mock_transport)

    p.initializing()
    assert not mock_transport.send.called
    p.create_message()

    mock_transport.send.assert_called_once()

    p.create_message()

    assert mock_transport.send.call_count == 2
    calls = mock_transport.send.call_args_list
    assert calls[0][0][0] == calls[1][0][0]  # same destination
    assert calls[0][0][1] != calls[1][0][1]  # different message


def test_txnservice_subscribes_to_channel():
    """Check that the service subscribes to a queue with acknowledgements enabled."""
    p = workflows.services.sample_transaction.SampleTxn()
    mock_transport = mock.Mock()
    setattr(p, "_transport", mock_transport)

    p.initializing()

    mock_transport.subscribe.assert_called_once_with(
        mock.ANY, p.receive_message, acknowledgement=True
    )


def test_txnservice_crash_function_crashes_sometimes():
    """The crash should happen sometimes. Neither never nor always."""
    fn = workflows.services.sample_transaction.SampleTxn.crashpoint

    assert any(fn() for i in range(100))
    assert not all(fn() for i in range(100))


def setup_txnservice(crashpattern):
    """Common fixture for TXN tests"""
    p = workflows.services.sample_transaction.SampleTxn()
    mock_crash, mock_transport = mock.Mock(), mock.Mock()
    p.crashpoint = mock_crash
    mock_crash.side_effect = crashpattern
    mock_transport.transaction_begin.return_value = mock.sentinel.txn
    mock_transport.subscribe.return_value = mock.sentinel.subid
    setattr(p, "_transport", mock_transport)
    header = {"message-id": mock.sentinel.message_id}
    message = mock.sentinel.message

    p.initializing()
    p.receive_message(header, message)

    return p, mock_transport


def test_txnservice_uses_transactions_correctly():
    """The TXN service should consume messages in a transaction. When the service fails the messages must not be consumed."""
    p, mock_transport = setup_txnservice([True])

    mock_transport.transaction_begin.assert_called_once_with()
    mock_transport.ack.assert_not_called()
    mock_transport.send.assert_not_called()
    mock_transport.transaction_commit.assert_not_called()
    mock_transport.transaction_abort.assert_called_once_with(mock.sentinel.txn)

    p, mock_transport = setup_txnservice([False, True])

    mock_transport.transaction_begin.assert_called_once_with()
    mock_transport.ack.assert_called_once_with(
        mock.sentinel.message_id, mock.sentinel.subid, transaction=mock.sentinel.txn
    )
    mock_transport.send.assert_not_called()
    mock_transport.transaction_commit.assert_not_called()
    mock_transport.transaction_abort.assert_called_once_with(mock.sentinel.txn)

    p, mock_transport = setup_txnservice([False, False, True])

    mock_transport.transaction_begin.assert_called_once_with()
    mock_transport.ack.assert_called_once_with(
        mock.sentinel.message_id, mock.sentinel.subid, transaction=mock.sentinel.txn
    )
    mock_transport.send.assert_called_once_with(
        mock.ANY, mock.sentinel.message, transaction=mock.sentinel.txn
    )
    mock_transport.transaction_commit.assert_not_called()
    mock_transport.transaction_abort.assert_called_once_with(mock.sentinel.txn)

    p, mock_transport = setup_txnservice([False, False, False])

    mock_transport.transaction_begin.assert_called_once_with()
    mock_transport.ack.assert_called_once_with(
        mock.sentinel.message_id, mock.sentinel.subid, transaction=mock.sentinel.txn
    )
    mock_transport.send.assert_called_once_with(
        mock.ANY, mock.sentinel.message, transaction=mock.sentinel.txn
    )
    mock_transport.transaction_commit.assert_called_once_with(mock.sentinel.txn)
    mock_transport.transaction_abort.assert_not_called()



import configparser
import json
import threading
import time

import stomp
import workflows
from workflows.transport.common_transport import CommonTransport


class StompTransport(CommonTransport):
    """Abstraction layer for messaging infrastructure. Here we are using ActiveMQ
    with STOMP."""

    # Set some sensible defaults
    defaults = {
        "--stomp-host": "localhost",
        "--stomp-port": 61613,
        "--stomp-user": "admin",
        "--stomp-pass": "password",
        "--stomp-prfx": "",
    }
    # Effective configuration
    config = {}

    def __init__(self):
        self._connected = False
        self._namespace = ""
        self._idcounter = 0
        self._lock = threading.RLock()
        self._stomp_listener = stomp.listener.ConnectionListener()
        #   self._stomp_listener = stomp.PrintingListener()
        self._stomp_listener.on_message = self._on_message
        self._stomp_listener.on_before_message = lambda x, y: (x, y)

    def get_namespace(self):
        """Return the stomp namespace. This is a prefix used for all topic and
        queue names."""
        if self._namespace.endswith("."):
            return self._namespace[:-1]
        return self._namespace

    @classmethod
    def load_configuration_file(cls, filename):
        cfgparser = configparser.ConfigParser(allow_no_value=True)
        if not cfgparser.read(filename):
            raise workflows.Error(
                "Could not read from configuration file %s" % filename
            )
        for cfgoption, target in [
            ("host", "--stomp-host"),
            ("port", "--stomp-port"),
            ("password", "--stomp-pass"),
            ("username", "--stomp-user"),
            ("prefix", "--stomp-prfx"),
        ]:
            try:
                cls.defaults[target] = cfgparser.get("stomp", cfgoption)
            except configparser.NoOptionError:
                pass

    @classmethod
    def add_command_line_options(cls, parser):
        """function to inject command line parameters"""
        if "add_argument" in dir(parser):
            return cls.add_command_line_options_argparse(parser)
        else:
            return cls.add_command_line_options_optparse(parser)

    @classmethod
    def add_command_line_options_argparse(cls, argparser):
        """function to inject command line parameters into
        a Python ArgumentParser."""
        import argparse

        class SetParameter(argparse.Action):
            """callback object for ArgumentParser"""

            def __call__(self, parser, namespace, value, option_string=None):
                cls.config[option_string] = value
                if option_string == "--stomp-conf":
                    cls.load_configuration_file(value)

        argparser.add_argument(
            "--stomp-host",
            metavar="HOST",
            default=cls.defaults.get("--stomp-host"),
            help="Stomp broker address, default '%(default)s'",
            type=str,
            action=SetParameter,
        )
        argparser.add_argument(
            "--stomp-port",
            metavar="PORT",
            default=cls.defaults.get("--stomp-port"),
            help="Stomp broker port, default '%(default)s'",
            type=int,
            action=SetParameter,
        )
        argparser.add_argument(
            "--stomp-user",
            metavar="USER",
            default=cls.defaults.get("--stomp-user"),
            help="Stomp user, default '%(default)s'",
            type=str,
            action=SetParameter,
        )
        argparser.add_argument(
            "--stomp-pass",
            metavar="PASS",
            default=cls.defaults.get("--stomp-pass"),
            help="Stomp password",
            type=str,
            action=SetParameter,
        )
        argparser.add_argument(
            "--stomp-prfx",
            metavar="PRE",
            default=cls.defaults.get("--stomp-prfx"),
            help="Stomp namespace prefix, default '%(default)s'",
            type=str,
            action=SetParameter,
        )
        argparser.add_argument(
            "--stomp-conf",
            metavar="CNF",
            default=cls.defaults.get("--stomp-conf"),
            help="Stomp configuration file containing connection information, disables default values",
            type=str,
            action=SetParameter,
        )

    @classmethod
    def add_command_line_options_optparse(cls, optparser):
        """function to inject command line parameters into
        a Python OptionParser."""

        def set_parameter(option, opt, value, parser):
            """callback function for OptionParser"""
            cls.config[opt] = value
            if opt == "--stomp-conf":
                cls.load_configuration_file(value)

        optparser.add_option(
            "--stomp-host",
            metavar="HOST",
            default=cls.defaults.get("--stomp-host"),
            help="Stomp broker address, default '%default'",
            type="string",
            nargs=1,
            action="callback",
            callback=set_parameter,
        )
        optparser.add_option(
            "--stomp-port",
            metavar="PORT",
            default=cls.defaults.get("--stomp-port"),
            help="Stomp broker port, default '%default'",
            type="int",
            nargs=1,
            action="callback",
            callback=set_parameter,
        )
        optparser.add_option(
            "--stomp-user",
            metavar="USER",
            default=cls.defaults.get("--stomp-user"),
            help="Stomp user, default '%default'",
            type="string",
            nargs=1,
            action="callback",
            callback=set_parameter,
        )
        optparser.add_option(
            "--stomp-pass",
            metavar="PASS",
            default=cls.defaults.get("--stomp-pass"),
            help="Stomp password",
            type="string",
            nargs=1,
            action="callback",
            callback=set_parameter,
        )
        optparser.add_option(
            "--stomp-prfx",
            metavar="PRE",
            default=cls.defaults.get("--stomp-prfx"),
            help="Stomp namespace prefix, default '%default'",
            type="string",
            nargs=1,
            action="callback",
            callback=set_parameter,
        )
        optparser.add_option(
            "--stomp-conf",
            metavar="CNF",
            default=cls.defaults.get("--stomp-conf"),
            help="Stomp configuration file containing connection information, disables default values",
            type="string",
            nargs=1,
            action="callback",
            callback=set_parameter,
        )

    def connect(self):
        with self._lock:
            if self._connected:
                return True
            self._conn = stomp.Connection(
                [
                    (
                        self.config.get(
                            "--stomp-host", self.defaults.get("--stomp-host")
                        ),
                        int(
                            self.config.get(
                                "--stomp-port", self.defaults.get("--stomp-port")
                            )
                        ),
                    )
                ]
            )
            self._conn.set_listener("", self._stomp_listener)
            username = self.config.get(
                "--stomp-user", self.defaults.get("--stomp-user")
            )
            password = self.config.get(
                "--stomp-pass", self.defaults.get("--stomp-pass")
            )
            timeout = time.time() + 30
            connection_failure = threading.Event()
            handler_old = self._stomp_listener.on_disconnected
            self._stomp_listener.on_disconnected = lambda: connection_failure.set()
            try:
                if username or password:
                    self._conn.connect(username, password, wait=False)
                else:  # anonymous access
                    self._conn.connect(wait=False)
            except stomp.exception.ConnectFailedException:
                raise workflows.Disconnected(
                    "Could not initiate connection to stomp host"
                )
            while (
                time.time() < timeout
                and not self._conn.is_connected()
                and not connection_failure.is_set()
            ):
                time.sleep(0.02)
            self._stomp_listener.on_disconnected = handler_old
            if connection_failure.is_set() or not self._conn.is_connected():
                raise workflows.Disconnected(
                    "Could not initiate connection to stomp host"
                )
            self._namespace = self.config.get(
                "--stomp-prfx", self.defaults.get("--stomp-prfx")
            )
            if self._namespace and not self._namespace.endswith("."):
                self._namespace = self._namespace + "."
            self._connected = True
        return True

    def is_connected(self):
        """Return connection status"""
        self._connected = self._connected and self._conn.is_connected()
        return self._connected

    def disconnect(self):
        """Gracefully close connection to stomp server."""
        if self._connected:
            self._connected = False
            self._conn.disconnect()

    def broadcast_status(self, status):
        """Broadcast transient status information to all listeners"""
        self._broadcast(
            "transient.status",
            json.dumps(status),
            headers={"expires": str(int((15 + time.time()) * 1000))},
        )

    def _subscribe(self, sub_id, channel, callback, **kwargs):
        """Listen to a queue, notify via callback function.
        :param sub_id: ID for this subscription in the transport layer
        :param channel: Queue name to subscribe to
        :param callback: Function to be called when messages are received
        :param **kwargs: Further parameters for the transport layer. For example
          acknowledgement:  If true receipt of each message needs to be
                            acknowledged.
          exclusive:        Attempt to become exclusive subscriber to the queue.
          ignore_namespace: Do not apply namespace to the destination name
          priority:         Consumer priority, messages are sent to higher
                            priority consumers whenever possible.
          selector:         Only receive messages filtered by a selector. See
                            https://activemq.apache.org/activemq-message-properties.html
                            for potential filter criteria. Uses SQL 92 syntax.
          transformation:   Transform messages into different format. If set
                            to True, will use 'jms-object-json' formatting.
        """
        headers = {}
        if kwargs.get("exclusive"):
            headers["activemq.exclusive"] = "true"
        if kwargs.get("ignore_namespace"):
            destination = "/queue/" + channel
        else:
            destination = "/queue/" + self._namespace + channel
        if kwargs.get("priority"):
            headers["activemq.priority"] = kwargs["priority"]
        if kwargs.get("retroactive"):
            headers["activemq.retroactive"] = "true"
        if kwargs.get("selector"):
            headers["selector"] = kwargs["selector"]
        if kwargs.get("transformation"):
            if kwargs["transformation"] is True:
                headers["transformation"] = "jms-object-json"
            else:
                headers["transformation"] = kwargs["transformation"]
        if kwargs.get("acknowledgement"):
            ack = "client-individual"
        else:
            ack = "auto"

        self._conn.subscribe(destination, sub_id, headers=headers, ack=ack)

    def _subscribe_broadcast(self, sub_id, channel, callback, **kwargs):
        """Listen to a broadcast topic, notify via callback function.
        :param sub_id: ID for this subscription in the transport layer
        :param channel: Topic name to subscribe to
        :param callback: Function to be called when messages are received
        :param **kwargs: Further parameters for the transport layer. For example
          ignore_namespace: Do not apply namespace to the destination name
          retroactive:      Ask broker to send old messages if possible
          transformation:   Transform messages into different format. If set
                            to True, will use 'jms-object-json' formatting.
        """
        headers = {}
        if kwargs.get("ignore_namespace"):
            destination = "/topic/" + channel
        else:
            destination = "/topic/" + self._namespace + channel
        if kwargs.get("retroactive"):
            headers["activemq.retroactive"] = "true"
        if kwargs.get("transformation"):
            if kwargs["transformation"] is True:
                headers["transformation"] = "jms-object-json"
            else:
                headers["transformation"] = kwargs["transformation"]
        self._conn.subscribe(destination, sub_id, headers=headers)

    def _unsubscribe(self, subscription, **kwargs):
        """Stop listening to a queue or a broadcast
        :param subscription: Subscription ID to cancel
        """
        self._conn.unsubscribe(id=subscription)
        # Callback reference is kept as further messages may already have been received

    def _send(
        self, destination, message, headers=None, delay=None, expiration=None, **kwargs
    ):
        """Send a message to a queue.
        :param destination: Queue name to send to
        :param message: A string to be sent
        :param **kwargs: Further parameters for the transport layer. For example
          delay:            Delay transport of message by this many seconds
          expiration:       Optional expiration time, relative to sending time
          headers:          Optional dictionary of header entries
          ignore_namespace: Do not apply namespace to the destination name
          persistent:       Whether to mark messages as persistent, to be kept
                            between broker restarts. Default is 'true'.
          transaction:      Transaction ID if message should be part of a
                            transaction
        """
        if not headers:
            headers = {}
        if "persistent" not in headers:
            headers["persistent"] = "true"
        if delay:
            # The 'delay' mechanism is only supported when
            # schedulerSupport is enabled on the broker.
            headers["AMQ_SCHEDULED_DELAY"] = int(1000 * delay)
        if expiration:
            headers["expires"] = int((time.time() + expiration) * 1000)
        if kwargs.get("ignore_namespace"):
            destination = "/queue/" + destination
        else:
            destination = "/queue/" + self._namespace + destination
        try:
            self._conn.send(destination, message, headers=headers, **kwargs)
        except stomp.exception.NotConnectedException:
            self._connected = False
            raise workflows.Disconnected("No connection to stomp host")

    def _broadcast(
        self, destination, message, headers=None, delay=None, expiration=None, **kwargs
    ):
        """Broadcast a message.
        :param destination: Topic name to send to
        :param message: A string to be broadcast
        :param **kwargs: Further parameters for the transport layer. For example
          delay:            Delay transport of message by this many seconds
          expiration:       Optional expiration time, relative to sending time
          headers:          Optional dictionary of header entries
          ignore_namespace: Do not apply namespace to the destination name
          transaction:      Transaction ID if message should be part of a
                            transaction
        """
        if not headers:
            headers = {}
        if delay:
            headers["AMQ_SCHEDULED_DELAY"] = 1000 * delay
        if expiration:
            headers["expires"] = int((time.time() + expiration) * 1000)
        if kwargs.get("ignore_namespace"):
            destination = "/topic/" + destination
        else:
            destination = "/topic/" + self._namespace + destination
        try:
            self._conn.send(destination, message, headers=headers, **kwargs)
        except stomp.exception.NotConnectedException:
            self._connected = False
            raise

    def _transaction_begin(self, transaction_id, **kwargs):
        """Start a new transaction.
        :param transaction_id: ID for this transaction in the transport layer.
        :param **kwargs: Further parameters for the transport layer.
        """
        self._conn.begin(transaction=transaction_id)

    def _transaction_abort(self, transaction_id, **kwargs):
        """Abort a transaction and roll back all operations.
        :param transaction_id: ID of transaction to be aborted.
        :param **kwargs: Further parameters for the transport layer.
        """
        self._conn.abort(transaction_id)

    def _transaction_commit(self, transaction_id, **kwargs):
        """Commit a transaction.
        :param transaction_id: ID of transaction to be committed.
        :param **kwargs: Further parameters for the transport layer.
        """
        self._conn.commit(transaction_id)

    def _ack(self, message_id, subscription_id, **kwargs):
        """Acknowledge receipt of a message. This only makes sense when the
        'acknowledgement' flag was set for the relevant subscription.
        :param message_id: ID of the message to be acknowledged
        :param subscription: ID of the relevant subscriptiong
        :param **kwargs: Further parameters for the transport layer. For example
               transaction: Transaction ID if acknowledgement should be part of
                            a transaction
        """
        self._conn.ack(message_id, subscription_id, **kwargs)

    def _nack(self, message_id, subscription_id, **kwargs):
        """Reject receipt of a message. This only makes sense when the
        'acknowledgement' flag was set for the relevant subscription.
        :param message_id: ID of the message to be rejected
        :param subscription: ID of the relevant subscriptiong
        :param **kwargs: Further parameters for the transport layer. For example
               transaction: Transaction ID if rejection should be part of a
                            transaction
        """
        self._conn.nack(message_id, subscription_id, **kwargs)

    @staticmethod
    def _mangle_for_sending(message):
        """Function that any message will pass through before it being forwarded to
        the actual _send* functions.
        Stomp only deals with serialized strings, so serialize message as json.
        """
        return json.dumps(message)

    @staticmethod
    def _mangle_for_receiving(message):
        """Function that any message will pass through before it being forwarded to
        the receiving subscribed callback functions.
        This transport class only deals with serialized strings, so decode
        message from json. However anything can come into here, so catch any
        deserialization errors.
        """
        try:
            return json.loads(message)
        except (TypeError, ValueError):
            return message

    ## Stomp listener methods #####################################################

    def _on_message(self, headers, body):
        subscription_id = int(headers.get("subscription"))
        target_function = self.subscription_callback(subscription_id)
        if target_function:
            target_function(headers, body)
        else:
            raise workflows.Error(
                "Unhandled message {} {}".format(repr(headers), repr(body))
            )



import logging

import workflows


class CommonTransport:
    """A common transport class, containing e.g. the logic to manage
    subscriptions and transactions."""

    __callback_interceptor = None
    __subscriptions = {}
    __subscription_id = 0
    __transactions = set()
    __transaction_id = 0

    log = logging.getLogger("workflows.transport")

    #
    # -- High level communication calls ----------------------------------------
    #

    @staticmethod
    def connect():
        """Connect the transport class. This function must be overridden.
        :return: True-like value when connection successful,
                 False-like value otherwise."""
        return False

    @staticmethod
    def is_connected():
        """Returns the current connection status. This function must be overridden.
        :return: True-like value when connection is available,
                 False-like value otherwise."""
        return False

    @staticmethod
    def disconnect():
        """Gracefully disconnect the transport class. This function should be
        overridden."""

    def subscribe(self, channel, callback, **kwargs):
        """Listen to a queue, notify via callback function.
        :param channel: Queue name to subscribe to
        :param callback: Function to be called when messages are received.
                         The callback will pass two arguments, the header as a
                         dictionary structure, and the message.
        :param **kwargs: Further parameters for the transport layer. For example
               disable_mangling: Receive messages as unprocessed strings.
               exclusive: Attempt to become exclusive subscriber to the queue.
               acknowledgement: If true receipt of each message needs to be
                                acknowledged.
        :return: A unique subscription ID
        """
        self.__subscription_id += 1

        def mangled_callback(header, message):
            return callback(header, self._mangle_for_receiving(message))

        if "disable_mangling" in kwargs:
            if kwargs["disable_mangling"]:
                mangled_callback = callback  # noqa:F811
            del kwargs["disable_mangling"]
        self.__subscriptions[self.__subscription_id] = {
            "channel": channel,
            "callback": mangled_callback,
            "ack": kwargs.get("acknowledgement"),
            "unsubscribed": False,
        }
        self.log.debug("Subscribing to %s with ID %d", channel, self.__subscription_id)
        self._subscribe(self.__subscription_id, channel, mangled_callback, **kwargs)
        return self.__subscription_id

    def unsubscribe(self, subscription, drop_callback_reference=False, **kwargs):
        """Stop listening to a queue or a broadcast
        :param subscription: Subscription ID to cancel
        :param drop_callback_reference: Drop the reference to the registered
                                        callback function immediately. This
                                        means any buffered messages still in
                                        flight will not arrive at the intended
                                        destination and cause exceptions to be
                                        raised instead.
        :param **kwargs: Further parameters for the transport layer.
        """
        if subscription not in self.__subscriptions:
            raise workflows.Error("Attempting to unsubscribe unknown subscription")
        if self.__subscriptions[subscription]["unsubscribed"]:
            raise workflows.Error(
                "Attempting to unsubscribe already unsubscribed subscription"
            )
        self._unsubscribe(subscription, **kwargs)
        self.__subscriptions[subscription]["unsubscribed"] = True
        if drop_callback_reference:
            self.drop_callback_reference(subscription)

    def drop_callback_reference(self, subscription):
        """Drop reference to the callback function after unsubscribing.
        Any future messages arriving for that subscription will result in
        exceptions being raised.
        :param subscription: Subscription ID to delete callback reference for.
        """
        if subscription not in self.__subscriptions:
            raise workflows.Error(
                "Attempting to drop callback reference for unknown subscription"
            )
        if not self.__subscriptions[subscription]["unsubscribed"]:
            raise workflows.Error(
                "Attempting to drop callback reference for live subscription"
            )
        del self.__subscriptions[subscription]

    def subscribe_broadcast(self, channel, callback, **kwargs):
        """Listen to a broadcast topic, notify via callback function.
        :param channel: Topic name to subscribe to
        :param callback: Function to be called when messages are received.
                         The callback will pass two arguments, the header as a
                         dictionary structure, and the message.
        :param **kwargs: Further parameters for the transport layer. For example
               disable_mangling: Receive messages as unprocessed strings.
               retroactive: Ask broker to send old messages if possible
        :return: A unique subscription ID
        """
        self.__subscription_id += 1

        def mangled_callback(header, message):
            return callback(header, self._mangle_for_receiving(message))

        if "disable_mangling" in kwargs:
            if kwargs["disable_mangling"]:
                mangled_callback = callback  # noqa:F811
            del kwargs["disable_mangling"]
        self.__subscriptions[self.__subscription_id] = {
            "channel": channel,
            "callback": mangled_callback,
            "ack": False,
            "unsubscribed": False,
        }
        self.log.debug(
            "Subscribing to broadcasts on %s with ID %d",
            channel,
            self.__subscription_id,
        )
        self._subscribe_broadcast(
            self.__subscription_id, channel, mangled_callback, **kwargs
        )
        return self.__subscription_id

    def subscription_callback(self, subscription):
        """Retrieve the callback function for a subscription. Raise a
        workflows.Error if the subscription does not exist.
        All transport callbacks can be intercepted by setting an
        interceptor function with subscription_callback_intercept().
        :param subscription: Subscription ID to look up
        :return: Callback function
        """
        subscription_record = self.__subscriptions.get(subscription)
        if not subscription_record:
            raise workflows.Error("Attempting to callback on unknown subscription")
        callback = subscription_record["callback"]
        if self.__callback_interceptor:
            return self.__callback_interceptor(callback)
        return callback

    def subscription_callback_set_intercept(self, interceptor):
        """Set a function to intercept all callbacks. This is useful to, for
        example, keep a thread barrier between the transport related functions
        and processing functions.
        :param interceptor: A function that takes the original callback function
                            and returns a modified callback function. Or None to
                            disable interception.
        """
        self.__callback_interceptor = interceptor

    def send(self, destination, message, **kwargs):
        """Send a message to a queue.
        :param destination: Queue name to send to
        :param message: Either a string or a serializable object to be sent
        :param **kwargs: Further parameters for the transport layer. For example
               delay: Delay transport of message by this many seconds
               headers: Optional dictionary of header entries
               expiration: Optional expiration time, relative to sending time
               transaction: Transaction ID if message should be part of a
                            transaction
        """
        message = self._mangle_for_sending(message)
        self._send(destination, message, **kwargs)

    def raw_send(self, destination, message, **kwargs):
        """Send a raw (unmangled) message to a queue.
        This may cause errors if the receiver expects a mangled message.
        :param destination: Queue name to send to
        :param message: Either a string or a serializable object to be sent
        :param **kwargs: Further parameters for the transport layer. For example
               delay: Delay transport of message by this many seconds
               headers: Optional dictionary of header entries
               expiration: Optional expiration time, relative to sending time
               transaction: Transaction ID if message should be part of a
                            transaction
        """
        self._send(destination, message, **kwargs)

    def broadcast(self, destination, message, **kwargs):
        """Broadcast a message.
        :param destination: Topic name to send to
        :param message: Either a string or a serializable object to be sent
        :param **kwargs: Further parameters for the transport layer. For example
               delay: Delay transport of message by this many seconds
               headers: Optional dictionary of header entries
               expiration: Optional expiration time, relative to sending time
               transaction: Transaction ID if message should be part of a
                            transaction
        """
        message = self._mangle_for_sending(message)
        self._broadcast(destination, message, **kwargs)

    def raw_broadcast(self, destination, message, **kwargs):
        """Broadcast a raw (unmangled) message.
        This may cause errors if the receiver expects a mangled message.
        :param destination: Topic name to send to
        :param message: Either a string or a serializable object to be sent
        :param **kwargs: Further parameters for the transport layer. For example
               delay: Delay transport of message by this many seconds
               headers: Optional dictionary of header entries
               expiration: Optional expiration time, relative to sending time
               transaction: Transaction ID if message should be part of a
                            transaction
        """
        self._broadcast(destination, message, **kwargs)

    def ack(self, message, subscription_id=None, **kwargs):
        """Acknowledge receipt of a message. This only makes sense when the
        'acknowledgement' flag was set for the relevant subscription.
        :param message: ID of the message to be acknowledged, OR a dictionary
                        containing a field 'message-id'.
        :param subscription_id: ID of the associated subscription. Optional when
                                a dictionary is passed as first parameter and
                                that dictionary contains field 'subscription'.
        :param **kwargs: Further parameters for the transport layer. For example
               transaction: Transaction ID if acknowledgement should be part of
                            a transaction
        """
        if isinstance(message, dict):
            message_id = message.get("message-id")
            if not subscription_id:
                subscription_id = message.get("subscription")
        else:
            message_id = message
        if not message_id:
            raise workflows.Error("Cannot acknowledge message without " + "message ID")
        if not subscription_id:
            raise workflows.Error(
                "Cannot acknowledge message without " + "subscription ID"
            )
        self.log.debug(
            "Acknowledging message %s on subscription %s", message_id, subscription_id
        )
        self._ack(message_id, subscription_id, **kwargs)

    def nack(self, message, subscription_id=None, **kwargs):
        """Reject receipt of a message. This only makes sense when the
        'acknowledgement' flag was set for the relevant subscription.
        :param message: ID of the message to be rejected, OR a dictionary
                        containing a field 'message-id'.
        :param subscription_id: ID of the associated subscription. Optional when
                                a dictionary is passed as first parameter and
                                that dictionary contains field 'subscription'.
        :param **kwargs: Further parameters for the transport layer. For example
               transaction: Transaction ID if rejection should be part of a
                            transaction
        """
        if isinstance(message, dict):
            message_id = message.get("message-id")
            if not subscription_id:
                subscription_id = message.get("subscription")
        else:
            message_id = message
        if not message_id:
            raise workflows.Error("Cannot reject message without " + "message ID")
        if not subscription_id:
            raise workflows.Error("Cannot reject message without " + "subscription ID")
        self.log.debug(
            "Rejecting message %s on subscription %s", message_id, subscription_id
        )
        self._nack(message_id, subscription_id, **kwargs)

    def transaction_begin(self, **kwargs):
        """Start a new transaction.
        :param **kwargs: Further parameters for the transport layer. For example
        :return: A transaction ID that can be passed to other functions.
        """
        self.__transaction_id += 1
        self.__transactions.add(self.__transaction_id)
        self.log.debug("Starting transaction with ID %d", self.__subscription_id)
        self._transaction_begin(self.__transaction_id, **kwargs)
        return self.__transaction_id

    def transaction_abort(self, transaction_id, **kwargs):
        """Abort a transaction and roll back all operations.
        :param transaction_id: ID of transaction to be aborted.
        :param **kwargs: Further parameters for the transport layer.
        """
        if transaction_id not in self.__transactions:
            raise workflows.Error("Attempting to abort unknown transaction")
        self.log.debug("Aborting transaction %s", transaction_id)
        self.__transactions.remove(transaction_id)
        self._transaction_abort(transaction_id, **kwargs)

    def transaction_commit(self, transaction_id, **kwargs):
        """Commit a transaction.
        :param transaction_id: ID of transaction to be committed.
        :param **kwargs: Further parameters for the transport layer.
        """
        if transaction_id not in self.__transactions:
            raise workflows.Error("Attempting to commit unknown transaction")
        self.log.debug("Committing transaction %s", transaction_id)
        self.__transactions.remove(transaction_id)
        self._transaction_commit(transaction_id, **kwargs)

    #
    # -- Low level communication calls to be implemented by subclass -----------
    #

    @staticmethod
    def _subscribe(sub_id, channel, callback, **kwargs):
        """Listen to a queue, notify via callback function.
        :param sub_id: ID for this subscription in the transport layer
        :param channel: Queue name to subscribe to
        :param callback: Function to be called when messages are received
        :param **kwargs: Further parameters for the transport layer. For example
               exclusive: Attempt to become exclusive subscriber to the queue.
               acknowledgement: If true receipt of each message needs to be
                                acknowledged.
        """
        raise NotImplementedError("Transport interface not implemented")

    @staticmethod
    def _subscribe_broadcast(sub_id, channel, callback, **kwargs):
        """Listen to a broadcast topic, notify via callback function.
        :param sub_id: ID for this subscription in the transport layer
        :param channel: Topic name to subscribe to
        :param callback: Function to be called when messages are received
        :param **kwargs: Further parameters for the transport layer. For example
               retroactive: Ask broker to send old messages if possible
        """
        raise NotImplementedError("Transport interface not implemented")

    @staticmethod
    def _unsubscribe(sub_id):
        """Stop listening to a queue or a broadcast
        :param sub_id: ID for this subscription in the transport layer
        """
        raise NotImplementedError("Transport interface not implemented")

    @staticmethod
    def _send(destination, message, **kwargs):
        """Send a message to a queue.
        :param destination: Queue name to send to
        :param message: A string to be sent
        :param **kwargs: Further parameters for the transport layer. For example
               headers: Optional dictionary of header entries
               expiration: Optional expiration time, relative to sending time
               transaction: Transaction ID if message should be part of a
                            transaction
        """
        raise NotImplementedError("Transport interface not implemented")

    @staticmethod
    def _broadcast(destination, message, **kwargs):
        """Broadcast a message.
        :param destination: Topic name to send to
        :param message: A string to be broadcast
        :param **kwargs: Further parameters for the transport layer. For example
               headers: Optional dictionary of header entries
               expiration: Optional expiration time, relative to sending time
               transaction: Transaction ID if message should be part of a
                            transaction
        """
        raise NotImplementedError("Transport interface not implemented")

    @staticmethod
    def _ack(message_id, subscription_id, **kwargs):
        """Acknowledge receipt of a message. This only makes sense when the
        'acknowledgement' flag was set for the relevant subscription.
        :param message_id: ID of the message to be acknowledged.
        :param subscription_id: ID of the associated subscription.
        :param **kwargs: Further parameters for the transport layer. For example
               transaction: Transaction ID if acknowledgement should be part of
                            a transaction
        """
        raise NotImplementedError("Transport interface not implemented")

    @staticmethod
    def _nack(message_id, subscription_id, **kwargs):
        """Reject receipt of a message. This only makes sense when the
        'acknowledgement' flag was set for the relevant subscription.
        :param message_id: ID of the message to be rejected.
        :param subscription_id: ID of the associated subscription.
        :param **kwargs: Further parameters for the transport layer. For example
               transaction: Transaction ID if rejection should be part of a
                            transaction
        """
        raise NotImplementedError("Transport interface not implemented")

    @staticmethod
    def _transaction_begin(transaction_id, **kwargs):
        """Start a new transaction.
        :param transaction_id: ID for this transaction in the transport layer.
        :param **kwargs: Further parameters for the transport layer.
        """
        raise NotImplementedError("Transport interface not implemented")

    @staticmethod
    def _transaction_abort(transaction_id, **kwargs):
        """Abort a transaction and roll back all operations.
        :param transaction_id: ID of transaction to be aborted.
        :param **kwargs: Further parameters for the transport layer.
        """
        raise NotImplementedError("Transport interface not implemented")

    @staticmethod
    def _transaction_commit(transaction_id, **kwargs):
        """Commit a transaction.
        :param transaction_id: ID of transaction to be committed.
        :param **kwargs: Further parameters for the transport layer.
        """
        raise NotImplementedError("Transport interface not implemented")

    #
    # -- Internal message mangling functions -----------------------------------
    #

    # Some transport mechanisms will not be able to work with arbitrary objects,
    # so these functions are used to prepare a message for sending/receiving.
    # The canonical example is serialization/deserialization, see stomp_transport

    @staticmethod
    def _mangle_for_sending(message):
        """Function that any message will pass through before it being forwarded to
        the actual _send* functions."""
        return message

    @staticmethod
    def _mangle_for_receiving(message):
        """Function that any message will pass through before it being forwarded to
        the receiving subscribed callback functions."""
        return message



from setuptools import find_packages, setup

with open("README.rst") as readme_file:
    readme = readme_file.read()

setup(
    name="workflows",
    description="Data processing in distributed environments",
    long_description=readme,
    url="https://github.com/DiamondLightSource/python-workflows",
    author="Markus Gerstel",
    author_email="scientificsoftware@diamond.ac.uk",
    download_url="https://github.com/DiamondLightSource/python-workflows/releases",
    version="2.2",
    install_requires=["setuptools", "stomp.py"],
    python_requires=">=3.6",
    packages=find_packages(),
    license="BSD",
    entry_points={
        "console_scripts": [
            "workflows.validate_recipe = workflows.recipe.validate:main"
        ],
        "libtbx.dispatcher.script": [
            "workflows.validate_recipe = workflows.validate_recipe"
        ],
        "libtbx.precommit": ["workflows = workflows"],
        "workflows.services": [
            "SampleConsumer = workflows.services.sample_consumer:SampleConsumer",
            "SampleProducer = workflows.services.sample_producer:SampleProducer",
            "SampleTxn = workflows.services.sample_transaction:SampleTxn",
            "SampleTxnProducer = workflows.services.sample_transaction:SampleTxnProducer",
        ],
        "workflows.transport": [
            "StompTransport = workflows.transport.stomp_transport:StompTransport"
        ],
    },
    tests_require=["pytest"],
    zip_safe=False,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



__version__ = "2.2"


def version():
    """Returns the version number of the installed workflows package."""
    return __version__


class Error(Exception):
    """Common class for exceptions deliberately raised by workflows package."""


class Disconnected(Error):
    """Indicates the connection could not be established or has been lost."""



import time

from workflows.services.common_service import CommonService


class SampleProducer(CommonService):
    """An example service building on top of the workflow.services architecture,
    demonstrating how this architecture can be used.
    This service generates messages into a queue."""

    # Human readable service name
    _service_name = "Message Producer"

    # Logger name
    _logger_name = "workflows.service.sample_producer"

    counter = 0

    def initializing(self):
        """Service initialization. This function is run before any commands are
        received from the frontend. This is the place to request channel
        subscriptions with the messaging layer, and register callbacks.
        This function can be overridden by specific service implementations."""
        self.log.info("Starting message producer")
        self._register_idle(3, self.create_message)

    def create_message(self):
        """Create and send a unique message for this service."""
        self.counter += 1
        self.log.info("Sending message #%d", self.counter)
        self._transport.send(
            "transient.destination",
            "Message #%d\n++++++++Produced@ %f"
            % (self.counter, (time.time() % 1000) * 1000),
        )



import contextlib
import enum
import itertools
import logging
import queue
import threading

import workflows
import workflows.logging


class Status(enum.Enum):
    """
    Internal service status codes
    ---------------------------------------------
    These codes will be sent to the frontend to indicate the current state of
    the main loop regardless of the status text, which can be set freely by
    the specific service.
    """

    # The state transitions are: (see definition of CommonService.start() below)
    #  constructor() -> NEW
    #            NEW -> start() being called -> STARTING
    #       STARTING -> self.initializing() -> IDLE
    #           IDLE -> wait for messages on command queue -> PROCESSING
    #              \--> optionally: idle timer elapsed -> TIMER
    #     PROCESSING -> process command -> IDLE
    #              \--> shutdown command received -> SHUTDOWN
    #          TIMER -> process event -> IDLE
    #       SHUTDOWN -> self.in_shutdown() -> END
    #  unhandled exception -> ERROR

    NEW = (0, "constructing")
    STARTING = (1, "starting")
    IDLE = (2, "idle")
    TIMER = (3, "timer event")
    PROCESSING = (4, "processing")
    SHUTDOWN = (5, "shutting down")
    END = (6, "shutdown")
    ERROR = (7, "error")

    # Extra states that are not used by services themselves but may be used
    # externally:

    NONE = (-1, "no service loaded")  # Node has no service instance loaded
    TEARDOWN = (-2, "shutdown")  # Node is shutting down

    def __init__(self, intval, description):
        """
        Each status is defined as a tuple of a unique integer value and a
        descriptive string. These are available via enum properties
        """
        self.intval = intval
        self.description = description


class Priority(enum.IntEnum):
    """
    Priorities for the service-internal priority queue. This ensures that eg.
    frontend commands are always processed before timer events.
    """

    COMMAND = 1
    TIMER = 2
    TRANSPORT = 3
    IDLE = 4


class CommonService:
    """
    Base class for workflow services. A service is a piece of software that runs
    in an isolated environment, communicating only via pipes with the outside
    world. Units of work are injected via a pipe. Results, status and log
    messages, etc. are written out via a pipe. Any task can be encapsulated
    as a service, for example a service that counts spots on an image passed
    as a filename, and returns the number of counts.

    To instantiate a service two Pipe-like objects should be passed to the
    constructors, one to communicate from the service to the frontend, one to
    communicate from the frontend to the service.
    """

    # Human readable service name -----------------------------------------------

    _service_name = "unnamed service"

    # Logger name ---------------------------------------------------------------

    _logger_name = "workflows.service"  # The logger can be accessed via self.log

    # Overrideable functions ----------------------------------------------------

    def initializing(self):
        """Service initialization. This function is run before any commands are
        received from the frontend. This is the place to request channel
        subscriptions with the messaging layer, and register callbacks.
        This function can be overridden by specific service implementations."""
        pass

    def in_shutdown(self):
        """Service shutdown. This function is run before the service is terminated.
        No more commands are received, but communications can still be sent.
        This function can be overridden by specific service implementations."""
        pass

    SERVICE_STATUS_NEW = Status.NEW.intval
    SERVICE_STATUS_STARTING = Status.STARTING.intval
    SERVICE_STATUS_IDLE = Status.IDLE.intval
    SERVICE_STATUS_TIMER = Status.TIMER.intval
    SERVICE_STATUS_PROCESSING = Status.PROCESSING.intval
    SERVICE_STATUS_SHUTDOWN = Status.SHUTDOWN.intval
    SERVICE_STATUS_END = Status.END.intval
    SERVICE_STATUS_ERROR = Status.ERROR.intval
    SERVICE_STATUS_NONE = Status.NONE.intval
    SERVICE_STATUS_TEARDOWN = Status.TEARDOWN.intval

    # Number to short string conversion

    human_readable_state = {e.intval: e.description for e in Status}  # lgtm

    # Default logging level for log messages from this service

    log_verbosity = logging.INFO

    # Any keyword arguments set on service invocation

    start_kwargs = {}

    # Not so overrideable functions ---------------------------------------------

    def __init__(self, *args, **kwargs):
        """Service constructor. Parameters include optional references to two
        pipes: frontend= for messages from the service to the frontend,
        and commands= for messages from the frontend to the service.
        A dictionary can optionally be passed with environment=, which is then
        available to the service during runtime."""
        self.__pipe_frontend = None
        self.__pipe_commands = None
        self._environment = kwargs.get("environment", {})
        self._transport = None
        self.__callback_register = {}
        self.__log_extensions = []
        self.__service_status = None
        self.__shutdown = False
        self.__update_service_status(self.SERVICE_STATUS_NEW)
        self.__queue = queue.PriorityQueue()
        self._idle_callback = None
        self._idle_time = None

        # Logger will be overwritten in start() function
        self.log = logging.getLogger(self._logger_name)

    def __send_to_frontend(self, data_structure):
        """Put a message in the pipe for the frontend."""
        if self.__pipe_frontend:
            self.__pipe_frontend.send(data_structure)

    @property
    def transport(self):
        return self._transport

    @transport.setter
    def transport(self, value):
        if self._transport:
            raise AttributeError("Transport already defined")
        self._transport = value

    def start_transport(self):
        """If a transport object has been defined then connect it now."""
        if self.transport:
            if self.transport.connect():
                self.log.debug("Service successfully connected to transport layer")
            else:
                raise RuntimeError("Service could not connect to transport layer")
            # direct all transport callbacks into the main queue
            self._transport_interceptor_counter = itertools.count()
            self.transport.subscription_callback_set_intercept(
                self._transport_interceptor
            )
        else:
            self.log.debug("No transport layer defined for service. Skipping.")

    def _transport_interceptor(self, callback):
        """Takes a callback function and returns a function that takes headers and
        messages and places them on the main service queue."""

        def add_item_to_queue(header, message):
            queue_item = (
                Priority.TRANSPORT,
                next(
                    self._transport_interceptor_counter
                ),  # insertion sequence to keep messages in order
                (callback, header, message),
            )
            self.__queue.put(
                queue_item
            )  # Block incoming transport until insertion completes

        return add_item_to_queue

    def connect(self, frontend=None, commands=None):
        """Inject pipes connecting the service to the frontend. Two arguments are
        supported: frontend= for messages from the service to the frontend,
        and commands= for messages from the frontend to the service.
        The injection should happen before the service is started, otherwise the
        underlying file descriptor references may not be handled correctly."""
        if frontend:
            self.__pipe_frontend = frontend
            self.__send_service_status_to_frontend()
        if commands:
            self.__pipe_commands = commands

    @contextlib.contextmanager
    def extend_log(self, field, value):
        """A context wherein a specified extra field in log messages is populated
        with a fixed value. This affects all log messages within the context."""
        self.__log_extensions.append((field, value))
        try:
            yield
        except Exception as e:
            setattr(e, "workflows_log_" + field, value)
            raise
        finally:
            self.__log_extensions.remove((field, value))

    def __command_queue_listener(self):
        """Function to continuously retrieve data from the frontend. Commands are
        sent to the central priority queue. If the pipe from the frontend is
        closed the service shutdown is initiated. Check every second if service
        has shut down, then terminate.
        This function is run by a separate daemon thread, which is started by
        the __start_command_queue_listener function.
        """
        self.log.debug("Queue listener thread started")
        counter = itertools.count()  # insertion sequence to keep messages in order
        while not self.__shutdown:
            if self.__pipe_commands.poll(1):
                try:
                    message = self.__pipe_commands.recv()
                except EOFError:
                    # Pipe was closed by frontend. Shut down service.
                    self.__shutdown = True
                    self.log.error(
                        "Pipe closed by frontend, shutting down service", exc_info=True
                    )
                    break
                queue_item = (Priority.COMMAND, next(counter), message)
                try:
                    self.__queue.put(queue_item, True, 60)
                except queue.Full:
                    # If the message can't be stored within 60 seconds then the service is
                    # operating outside normal parameters. Try to shut it down.
                    self.__shutdown = True
                    self.log.error(
                        "Write to service priority queue failed, shutting down service",
                        exc_info=True,
                    )
                    break
        self.log.debug("Queue listener thread terminating")

    def __start_command_queue_listener(self):
        """Start the function __command_queue_listener in a separate thread. This
        function continuously listens to the pipe connected to the frontend.
        """
        thread_function = self.__command_queue_listener

        class QueueListenerThread(threading.Thread):
            def run(qltself):
                thread_function()

        assert not hasattr(self, "__queue_listener_thread")
        self.log.debug("Starting queue listener thread")
        self.__queue_listener_thread = QueueListenerThread()
        self.__queue_listener_thread.daemon = True
        self.__queue_listener_thread.name = "Command Queue Listener"
        self.__queue_listener_thread.start()

    def _log_send(self, logrecord):
        """Forward log records to the frontend."""
        for field, value in self.__log_extensions:
            setattr(logrecord, field, value)
        self.__send_to_frontend({"band": "log", "payload": logrecord})

    def _register(self, message_band, callback):
        """Register a callback function for a specific message band."""
        self.__callback_register[message_band] = callback

    def _register_idle(self, idle_time, callback):
        """Register a callback function that is run when idling for a given
        time span (in seconds)."""
        self._idle_callback = callback
        self._idle_time = idle_time

    def __update_service_status(self, statuscode):
        """Set the internal status of the service object, and notify frontend."""
        if self.__service_status != statuscode:
            self.__service_status = statuscode
            self.__send_service_status_to_frontend()

    def __send_service_status_to_frontend(self):
        """Actually send the internal status of the service object to the frontend."""
        self.__send_to_frontend(
            {"band": "status_update", "statuscode": self.__service_status}
        )

    def get_name(self):
        """Get the name for this service."""
        return self._service_name

    def _set_name(self, name):
        """Set a new name for this service, and notify the frontend accordingly."""
        self._service_name = name
        self.__send_to_frontend({"band": "set_name", "name": self._service_name})

    def _request_termination(self):
        """Terminate the service from the frontend side"""
        self.__send_to_frontend({"band": "request_termination"})

    def _shutdown(self):
        """Terminate the service from the service side."""
        self.__shutdown = True

    def initialize_logging(self):
        """Reset the logging for the service process. All logged messages are
        forwarded to the frontend. If any filtering is desired, then this must
        take place on the service side."""
        # Reset logging to pass logrecords into the queue to the frontend only.
        # Existing handlers may be broken as they were copied into a new process,
        # so should be discarded.
        for loggername in [None] + list(logging.Logger.manager.loggerDict.keys()):
            logger = logging.getLogger(loggername)
            while logger.handlers:
                logger.removeHandler(logger.handlers[0])

        # Re-enable logging to console
        root_logger = logging.getLogger()

        # By default pass all warning (and higher) level messages to the frontend
        root_logger.setLevel(logging.WARN)
        root_logger.addHandler(workflows.logging.CallbackHandler(self._log_send))

        # Set up the service logger and pass all info (and higher) level messages
        # (or other level if set differently)
        self.log = logging.getLogger(self._logger_name)

        if self.start_kwargs.get("verbose_log"):
            self.log_verbosity = logging.DEBUG
        self.log.setLevel(self.log_verbosity)

        # Additionally, write all critical messages directly to console
        console = logging.StreamHandler()
        console.setLevel(logging.CRITICAL)
        root_logger.addHandler(console)

    def start(self, **kwargs):
        """Start listening to command queue, process commands in main loop,
        set status, etc...
        This function is most likely called by the frontend in a separate
        process."""

        # Keep a copy of keyword arguments for use in subclasses
        self.start_kwargs.update(kwargs)
        try:
            self.initialize_logging()

            self.__update_service_status(self.SERVICE_STATUS_STARTING)

            self.start_transport()

            self.initializing()
            self._register("command", self.__process_command)

            if self.__pipe_commands is None:
                # can only listen to commands if command queue is defined
                self.__shutdown = True
            else:
                # start listening to command queue in separate thread
                self.__start_command_queue_listener()

            while not self.__shutdown:  # main loop
                self.__update_service_status(self.SERVICE_STATUS_IDLE)

                if self._idle_time is None:
                    task = self.__queue.get()
                else:
                    try:
                        task = self.__queue.get(True, self._idle_time)
                        run_idle_task = False
                    except queue.Empty:
                        run_idle_task = True
                    if run_idle_task:
                        # run this outside the 'except' to avoid exception chaining
                        self.__update_service_status(self.SERVICE_STATUS_TIMER)
                        if self._idle_callback:
                            self._idle_callback()
                        continue

                self.__update_service_status(self.SERVICE_STATUS_PROCESSING)

                if task[0] == Priority.COMMAND:
                    message = task[2]
                    if message and "band" in message:
                        processor = self.__callback_register.get(message["band"])
                        if processor is None:
                            self.log.warning(
                                "received message on unregistered band\n%s", message
                            )
                        else:
                            processor(message.get("payload"))
                    else:
                        self.log.warning(
                            "received message without band information\n%s", message
                        )
                elif task[0] == Priority.TRANSPORT:
                    callback, header, message = task[2]
                    callback(header, message)
                else:
                    self.log.warning("Unknown item on main service queue\n%r", task)

        except KeyboardInterrupt:
            self.log.warning("Ctrl+C detected. Shutting down.")

        except Exception as e:
            self.process_uncaught_exception(e)
            self.__update_service_status(self.SERVICE_STATUS_ERROR)
            self.in_shutdown()
            return

        try:
            self.__update_service_status(self.SERVICE_STATUS_SHUTDOWN)
            self.in_shutdown()
            self.__update_service_status(self.SERVICE_STATUS_END)
        except Exception as e:
            self.process_uncaught_exception(e)
            self.__update_service_status(self.SERVICE_STATUS_ERROR)

    def process_uncaught_exception(self, e):
        """This is called to handle otherwise uncaught exceptions from the service.
        The service will terminate either way, but here we can do things such as
        gathering useful environment information and logging for posterity."""
        # Add information about the actual exception to the log message
        # This includes the file, line and piece of code causing the exception.
        # exc_info=True adds the full stack trace to the log message.
        (
            exc_file_fullpath,
            exc_file,
            exc_lineno,
            exc_func,
            exc_line,
        ) = workflows.logging.get_exception_source()
        added_information = {
            "workflows_exc_lineno": exc_lineno,
            "workflows_exc_funcName": exc_func,
            "workflows_exc_line": exc_line,
            "workflows_exc_pathname": exc_file_fullpath,
            "workflows_exc_filename": exc_file,
        }
        for field in filter(lambda x: x.startswith("workflows_log_"), dir(e)):
            added_information[field[14:]] = getattr(e, field, None)
        self.log.critical(
            "Unhandled service exception: %s", e, exc_info=True, extra=added_information
        )

    def __process_command(self, command):
        """Process an incoming command message from the frontend."""
        if command == Commands.SHUTDOWN:
            self.__shutdown = True


class Commands:
    """A list of command strings used for communicating with the frontend."""

    SHUTDOWN = "shutdown"
